import json
import re
from abc import abstractmethod

from jsvl.utils.message_list import ml
from jsvl.utils.util import converted_type, combine, reserved_key, regex_keys
from jsvl.validations.validation import Validation
import jsvl.models.schema as schema_model


class DocValidation(Validation):

    def __init__(self):
        super().__init__()
        self.full_path = None

    def run(self, key, schema, doc, path, index, doc_is_dynamic):

        if type(doc) is None:
            return

        if schema.get(key).can_bypass:
            return

        self.full_path = combine(path, key)
        self.validate(key, schema, doc, path, index, doc_is_dynamic)

    @abstractmethod
    def validate(self, key, schema, doc, path, index, doc_is_dynamic):
        pass


class ValidateUnknownKeys(DocValidation):

    def validate(self, key, schema, doc, path, index, doc_is_dynamic):
        obj = schema.get(key)

        if obj is not None and obj.can_bypass:
            return False

        if obj is None:
            self.create_error(ml.unknown_key(path))


class ValidateRequiredFields(DocValidation):

    def validate(self, key, schema, doc, path, index, doc_is_dynamic):
        obj = schema.get(key)

        if obj.is_required and key not in doc.keys():
            self.create_error(ml.missing_required_key(self.full_path))


class ValidateDataEquality(DocValidation):

    def validate(self, key, schema, doc, path, index, doc_is_dynamic):

        obj = schema.get(key)

        if obj.binding is not None:
            return

        doc_val = doc.get(key)

        if doc_val is not None:
            expected_type_list = obj.data_type.split("|")
            actual_type = converted_type(doc_val)
            if actual_type not in expected_type_list:
                expected_types = " or ".join(expected_type_list)
                self.create_error(ml.data_inequality(self.full_path, expected_types, actual_type))


class ValidateDocBinding(DocValidation):

    def validate(self, key, schema, doc, path, index, doc_is_dynamic):

        # if __bind_regex__ is defined then ignore the __bind__ because __bind_regex__ has higher precedence
        if schema.get(key).regex_binding is not None:
            return

        binding = schema.get(key).binding
        actual_value = doc.get(key)

        if binding is None or actual_value is None:
            return

        binder = schema_model.schema_doc.get(reserved_key.binder)
        expected_value = binder.get(binding).val
        expected_type = type(expected_value)
        actual_type = type(actual_value)

        if expected_type in [str, int, float, bool]:
            self.__validate_str_binding(expected_value, actual_value, expected_type, actual_type)

        elif expected_type is dict:
            if str(expected_value) != str(actual_value):
                self.create_error(ml.invalid_binding_object(self.full_path, expected_value))

        elif expected_type is list:

            if len(expected_value) == 0:
                if actual_type is not list or len(actual_value) > 0:
                    self.create_error(ml.empty_array_binding(self.full_path))

            elif str(actual_value) not in [str(item) for item in expected_value]:
                fields = json.dumps(expected_value, indent=4)
                self.create_error(ml.invalid_binding_array_item(self.full_path, binding, fields))

    def __validate_str_binding(self, expected_value, actual_value, expected_type, actual_type):

        if actual_type is not expected_type:
            self.create_error(ml.invalid_data_type_with_bind(self.full_path, converted_type(expected_type),
                                                             converted_type(actual_type)))

        elif len(str(expected_value)) == 0 and len(str(actual_value)) > 0:
            self.create_error(ml.empty_binding(self.full_path))

        elif actual_value != expected_value:
            self.create_error(ml.invalid_data_type_with_bind(self.full_path, expected_value, actual_value))


class ValidateDocRegexBinding(DocValidation):

    def validate(self, key, schema, doc, path, index, doc_is_dynamic):
        regex_pattern = schema.get(key).regex_binding
        regex_error = schema.get(key).regex_error_message
        actual_value = doc.get(key)

        if regex_pattern is None or actual_value is None:
            return

        regex_pattern = str(regex_pattern)
        regex_pattern = regex_pattern if regex_keys.get(regex_pattern) is None else regex_keys.get(regex_pattern)

        if not re.match(regex_pattern, str(actual_value)):
            self.create_error(ml.regex_binding_error(self.full_path, regex_error))


class ValidateDocTextCase(DocValidation):

    def validate(self, key, schema, doc, path, index, doc_is_dynamic):
        actual_value = doc.get(key)
        case = schema.get(key).case

        if case is None or actual_value is None:
            return

        if case == reserved_key.upper and not str(actual_value).isupper():
            self.create_error(ml.uppercase_error(self.full_path))

        elif case == reserved_key.lower and not str(actual_value).islower():
            self.create_error(ml.lowercase_error(self.full_path))

        elif case == reserved_key.title and not str(actual_value).istitle():
            self.create_error(ml.titlecase_error(self.full_path))


class ValidateTextSpace(DocValidation):

    def validate(self, key, schema, doc, path, index, doc_is_dynamic):
        actual_value = doc.get(key)
        allow_space = schema.get(key).allow_space

        if allow_space is not None and actual_value is not None:
            if not bool(allow_space) and " " in actual_value:
                self.create_error(ml.space_error(self.full_path))


# create validation set
doc_validation_set = {
    ValidateRequiredFields(),
    ValidateDataEquality(),
    ValidateDocBinding(),
    ValidateDocRegexBinding(),
    ValidateDocTextCase(),
    ValidateTextSpace()
}

validate_unknown_keys = ValidateUnknownKeys()




class MessageList:

    # ----------------------------------------------------------------------
    # schema validation errors, warnings, info, success
    # ----------------------------------------------------------------------

    def unmatched_data_type(self, expected_type, actual_type):
        return f"Expected {expected_type} type but found {actual_type} type."

    def min_length_item(self, expected_min_length):
        return f"The document should contain minimum {expected_min_length} item(s)."

    def max_length_item(self, expected_max_length):
        return f"The document should contain maximum {expected_max_length} item(s)."

    def validating_schema(self):
        return "Validating... schema."

    def validating_document(self):
        return "Validating... document."

    def schema_successfully_validated(self):
        return "Schema validated."

    def document_successfully_validated(self):
        return "Document validated."

    def validation_failed(self):
        return "Validation failed."

    def invalid_provided_schema(self):
        return "Provided schema is not valid."

    def invalid_provided_document(self):
        return "Provided document is not valid."

    def schema_root_object(self):
        return "In schema root object should be a json object."

    def schema_file_loaded(self, schema_file):
        return f"Loaded schema from: {schema_file}"

    def invalid_schema_file(self, schema_file):
        return f"Schema: {schema_file} is not a valid json file."

    def doc_file_loaded(self, doc_file):
        return f"Loaded document from: {doc_file}."

    def invalid_doc_file(self, doc_file):
        return f"Document: {doc_file} is not a valid json file."

    def file_not_found(self, dir):
        return f"No json file found in {dir}."

    def no_schema_found(self, schema_file_name):
        return f"{schema_file_name}.json not found."

    def unmatch_provided_schema_and_doc(self):
        return "if the provided document is a json content then the provided schema should also be json content or a single schema file."

    def empty_value_in_data_type(self, path):
        return f"{path} should not be empty."

    def found_invalid_data_type(self, path, invalid_data_type = ""):
        return f"{path} has invalid data type {invalid_data_type}. Check available data types here https://github.com/syedowaisali/json-validator-python?tab=readme-ov-file#support-data-types"

    def missing_binder_object(self):
        return "__binder__ object is missing from schema."

    def empty_schema(self):
        return "Schema couldn't be empty."

    def missing_binding(self, binding):
        return f"{binding} binding is missing."

    def invalid_binder_type(self):
        return "__binder__ type should be object."

    def invalid_regex_binding(self, path, regex_binding):
        return f"{path} has not a valid regex, {regex_binding}."

    def invalid_value_type(self, path, expected_type, actual_type):
        return f"{path} should be {expected_type} value, but found {actual_type} value."

    def key_support_with_string(self, path):
        return f"{path} can be used only with string data type."

    def key_support_with_number(self, path):
        return f"{path} can be used only with integer or float data type."

    def key_support_with_object_and_array(self, path):
        return f"{path} can be used only with object or object_array data type."

    def key_support_for_defaults(self, path):
        return f"{path} cannot be used in defaults. Check available keywords here https://github.com/syedowaisali/json-validator-python?tab=readme-ov-file#available-keywords"

    def invalid_min_max_value_or_length(self, min_value_path, max_value_path):
        return f"{min_value_path} should be less than {max_value_path} value."

    def negative_min_length(self, path):
        return f"{path} cannot be negative."

    def negative_or_zero_max_length(self, path):
        return f"{path} cannot be negative or 0."
    
    def invalid_text_case(self, path, case):
        return f"{path} has invalid text case {case}, Valid cases are __upper__, __lower__, and __title__. Check available keywords here https://github.com/syedowaisali/json-validator-python?tab=readme-ov-file#available-keywords"

    # def root_key_support(self):
    #    return "root object data type could be object, object_array, string_array, integer_array, float_array or bool_array data type."

    # ----------------------------------------------------------------------
    # document validation errors, warnings, info, success
    # ----------------------------------------------------------------------

    def empty_document_object(self):
        return "document shouldn't be empty object."

    def unknown_key(self, path):
        return f"{path} is unknown key, remove it or add it in schema."

    def missing_required_key(self, path):
        return f"{path} key is missing."

    def data_inequality(self, path, expected_types, actual_type):
        return f"{path} expect {expected_types} value, but found {actual_type} value."

    def invalid_data_type_with_bind(self, path, expected, actual):
        return f"{path} should be {expected}, but found {actual}."

    def empty_binding(self, path):
        return f"{path} should be empty."

    def empty_array_binding(self, path):
        return f"{path} should be empty array."

    def invalid_binding_object(self, path, expected):
        return f"{path} should be the same as the following object: {expected}"

    def invalid_binding_array_item(self, path, binding, array):
        return f"{path} should have one of the following value from {binding} array:\n \"{binding}\":{array}"

    def regex_binding_error(self, path, error=None):
        error = str(error) if error is not None else "has invalid value"
        return f"{path} {error}."

    def uppercase_error(self, path):
        return f"{path} should be upper case."

    def lowercase_error(self, path):
        return f"{path} should be lower case."

    def titlecase_error(self, path):
        return f"{path} should be title case."

    def space_error(self, path):
        return f"{path} cannot contain any space."

ml = MessageList()


import json

from jsvl.core.validator import validate
from jsvl.utils.message_list import ml
from jsvl.utils.util import data_type_cls, reserved_key
from jsvl.validations.doc_validations import ValidateUnknownKeys, ValidateRequiredFields, ValidateDataEquality, \
    ValidateDocBinding, ValidateDocRegexBinding, ValidateDocTextCase, ValidateTextSpace


class TestDocValidation:

    def test_validate_unknown_keys(self):
        schema = {
            "id": {},
            "location": {
                "__data_type__": "object"
            },
            "orders": {
                "__data_type__": "object_array",
                "origin": {
                    "__data_type__": "object",
                    "use": {}
                }
            }
        }

        doc = {
           "id": "12345",
            "name": "Michael",
            "age": 23,
            "location": {
                "address": {
                    "street": "11-B"
                }
            },
            "orders": [
                {
                    "origin": {
                        "source": "invy",
                        "use": "system"
                    }
                },
                {
                    "origin": {
                        "source": "sinvy"
                    }
                }
            ]
        }

        actual_logs = self.__validate(schema, doc, validation=ValidateUnknownKeys)
        expected_logs = [
            ml.unknown_key("name"),
            ml.unknown_key("age"),
            ml.unknown_key("location.address"),
            ml.unknown_key("orders[0].origin.source"),
            ml.unknown_key("orders[1].origin.source"),
        ]

        self.__assert_result(actual_logs, expected_logs, 5)

    def test_validate_required_fields(self):
        schema = {
            "id*": {},
            "address*": {
                "__data_type__": "object",
                "city*": {}
            },
            "orders*": {
                "__data_type__": "object_array",
                "id*": {},
                "title*": {},
                "detail": {},
                "location*": {
                    "__data_type__": "object",
                    "address*": {}
                }
            }
        }

        doc = {
            "style": "new",
            "orders": [
                {},
                {}
            ]
        }

        actual_logs = self.__validate(schema, doc, validation=ValidateRequiredFields)
        expected_logs = [
            ml.missing_required_key("id"),
            ml.missing_required_key("address"),
            ml.missing_required_key("orders[0].id"),
            ml.missing_required_key("orders[1].id"),
            ml.missing_required_key("orders[0].title"),
            ml.missing_required_key("orders[1].title"),
            ml.missing_required_key("orders[0].location"),
            ml.missing_required_key("orders[1].location"),
        ]

        self.__assert_result(actual_logs, expected_logs, 8)

    def test_validate_data_equality(self):
        schema = {
            "type": {},
            "info": {
                "__data_type__": "object",
                "serial": {
                    "__data_type__": "integer|float"
                },
                "can_use": {
                    "__data_type__": "bool"
                }
            },
            "records": {
                "__data_type__": "object_array",
                "id": {
                    "__data_type__": "integer",
                },
                "rating": {
                    "__data_type__": "float"
                }
            }
        }

        doc = {
            "type": 0,
            "info": {
                "serial": "one",
                "can_use": []
            },
            "records": [
                {
                    "id": True,
                    "rating": False
                },
                {
                    "id": 0.3,
                    "rating": "high"
                },
                {
                    "id": 2,
                    "rating": 3.0
                }
            ]
        }

        actual_logs = self.__validate(schema, doc, validation=ValidateDataEquality)
        expected_logs = [
            ml.data_inequality("type", data_type_cls.string, data_type_cls.integer),
            ml.data_inequality("info.serial", f"{data_type_cls.integer} or {data_type_cls.float}", data_type_cls.string),
            ml.data_inequality("info.can_use", data_type_cls.bool, data_type_cls.array),
            ml.data_inequality("records[0].id", data_type_cls.integer, data_type_cls.bool),
            ml.data_inequality("records[1].id", data_type_cls.integer, data_type_cls.float),
            ml.data_inequality("records[0].rating", data_type_cls.float, data_type_cls.bool),
            ml.data_inequality("records[1].rating", data_type_cls.float, data_type_cls.string),
        ]

        self.__assert_result(actual_logs, expected_logs, 7)

    def test_validate_binding(self):
        schema = {
            "country": {
                "__bind__": "countries"
            },
            "region": {
                "__bind__": "regions"
            },
            "board": {
                "__data_type__": "object",
                "score": {
                    "__bind__": "score"
                }
            },
            "model": {
                "__bind__": "structure"
            },
            "location": {
                "__bind__": "locations"
            },
            "pick": {
                "__bind__": "can_pick"
            },
            "__binder__": {
                "countries": "Pakistan",
                "score": 3,
                "structure": {
                    "floor": 4
                },
                "locations": ["KHI", "LHR", "ISB"],
                "regions": [],
                "can_pick": True
            }
        }

        doc = {
            "country": "Japan",
            "board": {
                "score": 1
            },
            "model": {
                "root": "low"
            },
            "location": "RWP",
            "region": "",
            "pick": 10
        }

        actual_logs = self.__validate(schema, doc, validation=ValidateDocBinding)
        fields = json.dumps(schema.get(reserved_key.binder).get("locations"), indent=4)
        expected_logs = [
            ml.invalid_data_type_with_bind("country", "Pakistan", "Japan"),
            ml.empty_array_binding("region"),
            ml.invalid_data_type_with_bind("board.score", 3, 1),
            ml.invalid_binding_object("model", "{'floor': 4}"),
            ml.invalid_data_type_with_bind("pick", "bool", "integer"),
            ml.invalid_binding_array_item("location", "locations", fields),
        ]

        self.__assert_result(actual_logs, expected_logs, 6)

    def test_validate_regex_bind(self):
        schema = {
            "id": {
                "__bind_regex__": "__numeric__",
                "__rem__": "should be contains only number"
            },
            "name": {
                "__data_type__": "object",
                "first_name": {
                    "__bind_regex__": "__alpha__",
                    "__rem__": "should contains only alphabets"
                },
                "last_name": {
                    "__bind_regex__": "__alpha_numeric__"
                }
            },
            "email": {
                "__bind_regex__": "__email__",
                "__rem__": "is not valid"
            },
            "parent_email": {
                "__bind_regex__": "__email__"
            },
            "ipv4": {
                "__bind_regex__": "__ipv4__",
            },
            "ipv6": {
                "__bind_regex__": "__ipv6__",
                "__rem__": "should be a valid IPV6"
            },
            "date": {
                "__bind_regex__": "^[0-9]{1,2}\\/[0-9]{1,2}\\/[0-9]{4}$",
                "__rem__": "should be like this DD/MM/YYYY"
            },
            "last_date": {
                "__bind_regex__": "^[0-9]{1,2}\\/[0-9]{1,2}\\/[0-9]{4}$"
            }
        }

        doc = {
            "id": "dash",
            "name": {
                "first_name": "Rocky 15",
                "last_name": "@"
            },
            "email": "rock@@gamil.com",
            "parent_email": "dev@gmail.com",
            "ipv4": "192.168.1",
            "ipv6": "3.223.4.5",
            "date": "05-01-2023",
            "last_date": "05/01/2023"
        }

        actual_logs = self.__validate(schema, doc, validation=ValidateDocRegexBinding)
        expected_logs = [
            ml.regex_binding_error("id", "should be contains only number"),
            ml.regex_binding_error("name.first_name", "should contains only alphabets"),
            ml.regex_binding_error("name.last_name"),
            ml.regex_binding_error("email", "is not valid"),
            ml.regex_binding_error("ipv4"),
            ml.regex_binding_error("ipv6", "should be a valid IPV6"),
            ml.regex_binding_error("date", "should be like this DD/MM/YYYY"),
        ]

        self.__assert_result(actual_logs, expected_logs, 7)

    def test_validate_text_case(self):
        schema = {
            "name": {
                "__case__": "__upper__"
            },
            "username": {
                "__case__": "__lower__"
            },
            "address": {
                "__data_type__": "object",
                "city": {
                    "__case__": "__title__"
                },
                "country": {
                    "__case__": "__title__"
                }
            }
        }

        doc = {
            "name": "owais",
            "username": "SyedOwaisAli",
            "address": {
                "city": "Karachi",
                "country": "pakistan"
            }
        }

        actual_logs = self.__validate(schema, doc, validation=ValidateDocTextCase)
        expected_logs = [
            ml.uppercase_error("name"),
            ml.lowercase_error("username"),
            ml.titlecase_error("address.country"),
        ]

        self.__assert_result(actual_logs, expected_logs, 3)

    def test_validate_text_space(self):
        schema = {
            "full_name": {
                "__allow_space__": False
            },
            "interest": {}
        }

        doc = {
            "full_name": "Bill G.",
            "interest": "Learn Engineering",
        }

        actual_logs = self.__validate(schema, doc, validation=ValidateTextSpace)
        expected_logs = [
            ml.space_error("full_name"),
        ]

        self.__assert_result(actual_logs, expected_logs, 1)

    def __assert_result(self, actual_logs: list, expected_logs: list, expected_size: int = -1):

        if expected_size > -1:
            assert len(actual_logs) == expected_size

        for log in expected_logs:
            assert log in actual_logs

    def __validate(self, schema, document, validation: object = None):

        output = validate(schema, document)[0].document_result

        if validation is not None:
            output = [result for result in output if type(result.validation) is validation]

        logs = [result.message for result in output]
        return logs



import json
import os.path
from glob import iglob
from typing import List

from ordered_set import OrderedSet

from jsvl.models.result import Output, Info, Error, Success, Warn

import jsvl.config as cfg
import jsvl.models.schema as schema_model
from jsvl.utils.message_list import ml
import jsvl.utils.util as util
from jsvl.validations.doc_validations import doc_validation_set, validate_unknown_keys
from jsvl.validations.schema_validations import schema_validation_set


def normalize_path(path: str, index: int = 0, doc_is_dynamic: bool = False) -> str:
    path = path.replace(cfg.root_object_path, "") if len(path) > len(cfg.root_object_path) else path
    path = path[1:] if path.startswith(".") else path
    return f"root[{index}]{path}" if doc_is_dynamic and "root" not in path else path


def apply_doc_unknown_keys_validation(key, schema, doc, path, index, doc_is_dynamic, output: OrderedSet):
    path = normalize_path(path, index, doc_is_dynamic)
    doc_value = doc.get(key)
    sch_value = schema.get(key) if schema is not None else None

    if key is None:
        for tgt_key in doc.keys():
            apply_doc_unknown_keys_validation(tgt_key, schema, doc, path, index, doc_is_dynamic, output)
        return

    if schema is not None and validate_unknown_keys.validate(key, schema, doc,
                                                  normalize_path(f"{path}.{key}", index, doc_is_dynamic), index,
                                                  doc_is_dynamic) is not None:
        return
    output.update(validate_unknown_keys.get_set())

    if type(doc_value) is dict:
        for child_key in doc_value.keys():
            apply_doc_unknown_keys_validation(child_key, sch_value, doc_value,
                                 normalize_path(f"{path}.{key}", index, doc_is_dynamic), index, doc_is_dynamic,
                                 output)

    if type(doc_value) is list:
        for i, item in enumerate(doc_value):
            if type(item) is dict:
                for child_key in item.keys():
                    apply_doc_unknown_keys_validation(child_key, sch_value, item, f"{path}.{key}[{i}]", index, doc_is_dynamic,
                                         output)

def apply_doc_validation(key, schema, doc, path, index, doc_is_dynamic, output: OrderedSet):
    path = f"root[{index}]{path}" if doc_is_dynamic and "root" not in path else path
    obj = schema.get(key)

    if key is None:
        for child_key in util.remove_reserved_keys(schema):
            apply_doc_validation(child_key, schema, doc, path, index, doc_is_dynamic, output)
        return

    #if type(doc) is list:
    #    for i, item in enumerate(doc):
    #        if type(item) is dict:
    #            apply_doc_validation(key, obj, item, f"{path}.{key}[{i}]", index, doc_is_dynamic, output)
    #    return

    for validation in doc_validation_set:
        if validation.run(key, schema, doc,
                          path, index,
                          doc_is_dynamic) is not None:
            return
        output.update(validation.get_set())

    if type(doc) is dict and obj.val_is_dict:

        for child_key in util.remove_reserved_keys(obj.val):

            doc_obj = doc.get(key)
            if type(doc_obj) is dict:
                apply_doc_validation(child_key, obj, doc_obj, f"{path}.{key}", index, doc_is_dynamic, output)

            if type(doc_obj) is list:
                for i, item in enumerate(doc_obj):
                    if type(item) is dict:
                        apply_doc_validation(child_key, obj, item, f"{path}.{key}[{i}]", index, doc_is_dynamic, output)


def apply_schema_validation(key, schema, path, output: OrderedSet):
    path = normalize_path(path)
    obj = schema.get(key)

    if key is None:
        for child_key in schema.keys():
            if child_key != util.reserved_key.binder:
                apply_schema_validation(child_key, schema, path, output)
        return

    for validation in schema_validation_set:
        if validation.validate(key, schema, normalize_path(f"{path}.{key}")) is not None:
            return

        output.update(validation.get_set())

    if type(obj.val) is dict:
        for child_key in util.remove_reserved_keys(obj.val):
            apply_schema_validation(child_key, obj.child_schema, f"{path}.{key}", output)


def prepare_schema(schema: dict, updated_schema: dict, key=None, defaults=None):
    if key is None:

        if util.reserved_key.data_type not in schema.keys():
            schema[util.reserved_key.data_type] = util.data_type_cls.object

        if util.reserved_key.defaults not in schema.keys():
            schema[util.reserved_key.defaults] = {
                util.reserved_key.allow_space: cfg.configs.get(cfg.allow_space),
                util.reserved_key.min_length: cfg.configs.get(cfg.min_length),
                util.reserved_key.max_length: cfg.configs.get(cfg.max_length),
                util.reserved_key.min_value: cfg.configs.get(cfg.min_value),
                util.reserved_key.max_value: cfg.configs.get(cfg.max_value),
                util.reserved_key.case: cfg.configs.get(cfg.case),
            }

        defaults = schema.get(util.reserved_key.defaults)

        if util.reserved_key.allow_space not in defaults:
            defaults[util.reserved_key.allow_space] = cfg.configs.get(cfg.allow_space)

        if util.reserved_key.min_length not in defaults:
            defaults[util.reserved_key.min_length] = cfg.configs.get(cfg.min_length)

        if util.reserved_key.max_length not in defaults:
            defaults[util.reserved_key.max_length] = cfg.configs.get(cfg.max_length)

        if util.reserved_key.min_value not in defaults:
            defaults[util.reserved_key.min_value] = cfg.configs.get(cfg.min_value)

        if util.reserved_key.max_value not in defaults:
            defaults[util.reserved_key.max_value] = cfg.configs.get(cfg.max_value)

        if util.reserved_key.case not in defaults:
            defaults[util.reserved_key.case] = cfg.configs.get(cfg.case)

        for s_key in schema.keys():
            prepare_schema(schema, updated_schema, s_key, defaults)
        return

    if type(schema.get(key)) is dict and key not in util.reserved_key.all_keys():
        field = schema[key].copy()

        data_type = util.data_type_cls.string if field.get(util.reserved_key.data_type) is None else field.get(
            util.reserved_key.data_type)
        is_required = True if key.endswith("*") else False
        bypass = True if key.startswith("~") else False

        field[util.reserved_key.data_type] = data_type
        field[util.reserved_key.required] = is_required
        field[util.reserved_key.bypass] = bypass

        if data_type == util.data_type_cls.string:
            allow_space = defaults.get(util.reserved_key.allow_space) if field.get(
                util.reserved_key.allow_space) is None else field.get(util.reserved_key.allow_space)
            min_length = defaults.get(util.reserved_key.min_length) if field.get(
                util.reserved_key.min_length) is None else field.get(util.reserved_key.min_length)
            max_length = defaults.get(util.reserved_key.max_length) if field.get(
                util.reserved_key.max_length) is None else field.get(util.reserved_key.max_length)
            case = defaults.get(util.reserved_key.case) if field.get(util.reserved_key.case) is None else field.get(
                util.reserved_key.case)

            field[util.reserved_key.allow_space] = allow_space
            field[util.reserved_key.min_length] = min_length
            field[util.reserved_key.max_length] = max_length
            field[util.reserved_key.case] = case

        if data_type == util.data_type_cls.integer:
            min_value = defaults.get(util.reserved_key.min_value) if field.get(
                util.reserved_key.min_value) is None else field.get(util.reserved_key.min_value)
            max_value = defaults.get(util.reserved_key.max_value) if field.get(
                util.reserved_key.max_value) is None else field.get(util.reserved_key.max_value)

            field[util.reserved_key.min_value] = min_value
            field[util.reserved_key.max_value] = max_value

        updated_key = key[:-1] if key.endswith("*") else key
        updated_key = updated_key[1:] if updated_key.startswith("~") else updated_key

        updated_schema[updated_key] = field
        if (key.startswith("~") or key.endswith("*")) and key in updated_schema.keys():
            del updated_schema[key]

        for child_key in schema.get(key).keys():
            prepare_schema(schema.get(key), updated_schema.get(updated_key), child_key, defaults)
    else:
        updated_schema[key] = schema.get(key)


def is_dynamic_document(document: list) -> bool:
    first_item_type = type(document[0])
    is_dynamic_array = False

    for item in document:
        if type(item) is not first_item_type:
            is_dynamic_array = True

    return is_dynamic_array


def check_root_data_type(schema: dict, expected_type: str, output_validation: OrderedSet):
    data_type = schema.get(util.reserved_key.data_type)

    if data_type != expected_type:
        output_validation.add(Error(ml.unmatched_data_type(expected_type, data_type)))


def validate_min_length(schema: dict, document: list, output_validation: OrderedSet):
    min_length = schema.get(util.reserved_key.min_length)
    if min_length is not None and len(document) < int(min_length):
        output_validation.add(Error(ml.min_length_item(min_length)))


def validate_max_length(schema: dict, document: list, output_validation: OrderedSet):
    max_length = schema.get(util.reserved_key.max_length)
    if max_length is not None and len(document) > int(max_length):
        output_validation.add(Error(ml.max_length_item(max_length)))


def run_doc_validations(schema: dict, document, doc_validation_output: OrderedSet):
    # target document comprises on a single object
    if type(document) is dict and len(document.keys()) > 0:
        apply_doc_validation(None, schema_model.schema_doc, document, "", 0, False,
                             doc_validation_output)

        apply_doc_unknown_keys_validation(None, schema_model.schema_doc, document, "", 0, False,
                             doc_validation_output)

    # target document comprises on a array
    elif len(document) > 0:

        # when document is dynamic
        if is_dynamic_document(document):
            check_root_data_type(schema, "array", doc_validation_output)

        # when document is not dynamic
        elif type(document[0]) is dict:
            for index, doc in enumerate(document):
                apply_doc_validation(None, schema_model.schema_doc, doc, "", index, True,
                                     doc_validation_output)
                apply_doc_unknown_keys_validation(None, schema_model.schema_doc, doc, "", index, True,
                                     doc_validation_output)

        elif type(document[0]) is str:
            check_root_data_type(schema, util.data_type_cls.string_array, doc_validation_output)

        elif type(document[0]) is int:
            check_root_data_type(schema, util.data_type_cls.integer_array, doc_validation_output)

        elif type(document[0]) is float:
            check_root_data_type(schema, util.data_type_cls.float_array, doc_validation_output)

        elif type(document[0]) is bool:
            check_root_data_type(schema, util.data_type_cls.bool_array, doc_validation_output)

        validate_min_length(schema, document, doc_validation_output)
        validate_max_length(schema, document, doc_validation_output)

    else:
        doc_validation_output.add(Error(ml.empty_document_object()))


def start_validation_process(schema: dict, document, schema_validation_output: OrderedSet,
                             doc_validation_output: OrderedSet):
    updated_schema = {}

    # remove * from keys and add __required__ key in field object
    # remove ~ from keys and add __bypass__ key in field object
    prepare_schema(schema, updated_schema)

    schema_model.schema_doc = {}
    for key in {key: val for (key, val) in updated_schema.items()}:
        schema_model.schema_doc[key] = schema_model.Schema(key, updated_schema[key])

    # adding separator and initial info
    schema_validation_output.add(Info(ml.validating_schema()))

    # validating schema first then document
    apply_schema_validation(None, schema_model.schema_doc, cfg.root_object_path, schema_validation_output)

    # if schema has no error then validate the target document with schema
    if len([result for result in schema_validation_output if type(result) is Error]) == 0:

        schema_validation_output.add(Success(ml.schema_successfully_validated()))
        schema_validation_output.add(Info(ml.validating_document()))

        run_doc_validations(schema, document, doc_validation_output)

        if len([result for result in doc_validation_output if type(result) is Error]) == 0:
            doc_validation_output.add(Success(ml.document_successfully_validated()))


def execute(schema, document, out: list):
    for validation in schema_validation_set:
        validation.init_log_set()

    for validation in doc_validation_set:
        validation.init_log_set()

    schema_validation_output = OrderedSet()
    doc_validation_output = OrderedSet()

    out.append(Output(schema_validation_output, doc_validation_output))

    schema_is_valid_json = util.is_valid_json(schema)
    schema_is_file = util.is_valid_file(schema)
    schema_is_dir = util.is_valid_dir(schema)

    if not schema_is_valid_json and not schema_is_file and not schema_is_dir:
        schema_validation_output.add(Error(ml.invalid_provided_schema()))
        return

    doc_is_valid_json = util.is_valid_json(document)
    doc_is_file = util.is_valid_file(document)
    doc_is_dir = util.is_valid_dir(document)

    if not doc_is_valid_json and not doc_is_file and not doc_is_dir:
        doc_validation_output.add(Error(ml.invalid_provided_document()))
        return

    list_of_schema = []

    if schema_is_valid_json:
        schema = util.get_as_json(schema)
        if type(schema) is not dict:
            schema_validation_output.add(Error(ml.schema_root_object()))
            return
        list_of_schema.append(schema)

    if schema_is_file:
        schema_validation_output.add(Info(ml.schema_file_loaded(schema)))
        file_content = util.read_file(schema)
        if not util.is_valid_json(file_content):
            schema_validation_output.add(Error(ml.invalid_schema_file(schema)))
            return

        schema = json.loads(file_content)
        list_of_schema.append(schema)

    if schema_is_dir:
        list_of_schema.clear()
        schema = schema if schema.endswith("/") else f"{schema}/"

        for file in iglob(f"{schema}**/*.json", recursive=True):
            list_of_schema.append(file)

    list_of_doc = []
    if doc_is_valid_json:
        document = util.get_as_json(document)
        list_of_doc.append(document)

    if doc_is_file:
        schema_validation_output.add(Info(ml.doc_file_loaded(document)))
        file_content = util.read_file(document)

        if not util.is_valid_json(file_content):
            doc_validation_output.add(Error(ml.invalid_doc_file(document)))
            return

        document = json.loads(file_content)
        list_of_doc.append(document)

    if doc_is_dir:
        list_of_doc.clear()
        document = document if document.endswith("/") else f"{document}/"

        for file in iglob(f"{document}**/*.json", recursive=True):
            list_of_doc.append(file)

        if len(list_of_doc) == 0:
            doc_validation_output.add(Info(ml.file_not_found(document)))
            return

        if len(list_of_schema) == 0:
            schema_validation_output.add(Error(ml.file_not_found(schema)))
            return

        for target_doc in list_of_doc:

            if type(schema) is dict:
                execute(schema, target_doc, out)

            elif schema == list_of_schema[0]:
                execute(schema, target_doc, out)
            else:
                schema_file_name = f"{os.path.splitext(os.path.basename(target_doc))[0]}{cfg.configs.get(cfg.schema_file_postfix)}"

                found_schema_file = None
                for schema_file_path in list_of_schema:
                    if f"/{schema_file_name}.json" in schema_file_path:
                        found_schema_file = schema_file_path
                        break

                if found_schema_file is not None:
                    execute(found_schema_file, target_doc, out)
                else:
                    schema_validation_output.add(Warn(ml.no_schema_found(schema_file_name)))

        return

    if type(document) is dict and (not schema_is_valid_json and not schema_is_file):
        schema_validation_output.add(Error(ml.unmatch_provided_schema_and_doc()))
        return

    if len(schema) == 0:
        schema_validation_output.add(Error(ml.empty_schema()))
        return

    start_validation_process(schema, document, schema_validation_output, doc_validation_output)


def validate(schema, document) -> List[Output]:
    output = []
    execute(schema, document, output)

    if cfg.configs.get(cfg.enable_output_logs):
        util.dump_log(output)

    return output



import argparse

import jsvl.config as cfg
from jsvl.utils.util import reserved_key


def setup_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="JSON Validator",
        description="JSON Validator is a feature-rich Python library designed to elevate the validation of JSON documents by offering an extensive schema-based validation system.",
    )

    parser.add_argument(
        "-s", "--schema",
        action="store",
        dest="schema",
        required=True,
        type=str,
        help="Provide a schema json, file or a directory path for validating the documents."
    )
    parser.add_argument(
        "-d", "--doc",
        dest="doc",
        required=True,
        type=str,
        help="Provide a document json, file or a directory path for validation."
    )
    parser.add_argument(
        "-csfp", "--change-schema-file-postfix",
        dest="schema_file_postfix",
        default="_schema",
        type=str,
        help="This argument will be used to change the schema file postfix name."
    )
    parser.add_argument(
        "--disable-tags",
        dest="enable_output_tags",
        action="store_false",
        help="Pass this flag to disable the output tags from logs."
    )
    parser.add_argument(
        "--plain-output",
        dest="formatted_output",
        action="store_false",
        help="Pass this flag to remove the formatting from logs."
    )
    parser.add_argument(
        "--disable-logs",
        dest="enable_output_logs",
        action="store_false",
        help="Pass this flag to disable all the logs."
    )
    parser.add_argument(
        "--show-validation-source",
        dest="enable_validation_source",
        action="store_true",
        help="Pass this flag to show the validation class name with logs."
    )
    parser.add_argument(
        "--tight-space",
        dest="allow_space",
        action="store_false",
        help="Pass this flag to disable the allow_space globally."
    )
    parser.add_argument(
        "-ml", "--min-length",
        dest="min_length",
        type=int,
        default=0,
        help="Set the minimum length globally, default is 0."
    )
    parser.add_argument(
        "-xl", "--max-length",
        dest="max_length",
        type=int,
        default=None,
        help="Set the maximum length globally, default is None."
    )
    parser.add_argument(
        "-mv", "--min-value",
        dest="min_value",
        type=float,
        default=0,
        help="Set the minimum value globally, default is 0."
    )
    parser.add_argument(
        "-xv", "--max-value",
        dest="max_value",
        type=float,
        default=None,
        help="Set the maximum value globally, default is None."
    )
    parser.add_argument(
        "-c", "--case",
        dest="case",
        type=str,
        choices=[reserved_key.upper, reserved_key.lower, reserved_key.title],
        default=None,
        help="Set the text constraints globally, default is None."
    )
    parser.add_argument(
        "-v", "--version",
        action="version",
        version="%(prog)s 1.0.7"
    )

    return parser


def parse(parser) -> argparse.Namespace:
    args = parser.parse_args()
    cfg.configs[cfg.schema_file_postfix] = args.schema_file_postfix
    cfg.configs[cfg.enable_output_tags] = args.enable_output_tags
    cfg.configs[cfg.formatted_output] = args.formatted_output
    cfg.configs[cfg.enable_output_logs] = args.enable_output_logs
    cfg.configs[cfg.enable_validation_source] = args.enable_validation_source
    cfg.configs[cfg.allow_space] = args.allow_space
    cfg.configs[cfg.min_length] = args.min_length
    cfg.configs[cfg.max_length] = args.max_length
    cfg.configs[cfg.min_value] = args.min_value
    cfg.configs[cfg.max_value] = args.max_value
    cfg.configs[cfg.case] = args.case

    return args



from abc import ABC

from jsvl.models.object_set import ObjectSet
from jsvl.models.result import Error, Warn, Info, Success


class Validation(ABC):

    def __init__(self):
        self.__results = ObjectSet()

    def create_error(self, message):
        self.__results.add(Error(message, self))

    def create_warn(self, message):
        self.__results.add(Warn(message, self))

    def create_info(self, message):
        self.__results.add(Info(message, self))

    def create_success(self, message):
        self.__results.add(Success(message, self))

    def init_log_set(self):
        self.__results.ini()

    def get_set(self):
        return self.__results.item_set()






from jsvl.core import validator
from jsvl.core.arg_parser import setup_arg_parser, parse


def run_validation():

    parser = setup_arg_parser()
    args = parse(parser)
    validator.validate(args.schema, args.doc)


# Press the green button in the gutter to run the script.
if __name__ == '__main__':
    run_validation()



from jsvl.utils.util import reserved_key


class Schema:

    def __init__(self, key, val):
        self.child_schema = {}
        self.val = val
        self.key = key

        self.is_required = False
        self.data_type = None
        self.allow_space = None
        self.min_value = None
        self.max_value = None
        self.min_length = None
        self.max_length = None
        self.case = None
        self.can_bypass = None
        self.binding = None
        self.regex_binding = None
        self.regex_error_message = None

        if type(val) is dict:
            self.is_required = val.get(reserved_key.required)
            self.data_type = val.get(reserved_key.data_type)
            self.allow_space = val.get(reserved_key.allow_space)
            self.min_value = val.get(reserved_key.min_value)
            self.max_value = val.get(reserved_key.max_value)
            self.min_length = val.get(reserved_key.min_length)
            self.max_length = val.get(reserved_key.max_length)
            self.case = val.get(reserved_key.case)
            self.can_bypass = val.get(reserved_key.bypass)
            self.binding = val.get(reserved_key.bind)
            self.regex_binding = val.get(reserved_key.bind_regex)
            self.regex_error_message = val.get(reserved_key.regex_error_message)

            keys = {k: v for (k, v) in val.items() if k not in reserved_key.all_keys().keys()}
            for new_key in keys:
                self.child_schema[new_key] = Schema(new_key, val.get(new_key))

    def get(self, key):
        return self.child_schema.get(key)

    def val_is_dict(self):
        return type(self.val) is dict


schema_doc = {}



from jsvl.core.validator import validate
from jsvl.utils.util import reserved_key, data_type_cls, is_valid_regex
from jsvl.utils.message_list import ml
from jsvl.validations.schema_validations import ValidateKeysCombinations, ValidateValueType, ValidateSchemaBindings, \
    ValidateDataType, ValidateMinMaxValue, ValidateTextCase


class TestSchemaValidation:

    def test_validate_data_type(self):
        schema = {
            "__data_type__": "number",
            "cart": {
                "location": {
                    "__data_type__": "s"
                },
                "primary": {
                    "__data_type__": ""
                },
                "secondary": {
                    "__data_type__": "j|string|9"
                }
            }
        }

        actual_logs = self.__validate(schema, validation=ValidateDataType)
        expected_logs = [
            ml.empty_value_in_data_type(f"cart.primary.{reserved_key.data_type}"),
            ml.found_invalid_data_type(f"cart.location.{reserved_key.data_type}", "s"),
            ml.found_invalid_data_type(f"cart.secondary.{reserved_key.data_type}", "j"),
            ml.found_invalid_data_type(f"cart.secondary.{reserved_key.data_type}", "9"),
            ml.found_invalid_data_type(reserved_key.data_type, "number"),
        ]

        self.__assert_result(actual_logs, expected_logs, 5)

    def test_validate_binder_object(self):
        schema = {
            "type": {
                reserved_key.bind: "roles"
            }
        }

        actual_logs = self.__validate(schema, validation=ValidateSchemaBindings)
        expected_logs = [ml.missing_binder_object()]

        self.__assert_result(actual_logs, expected_logs, 1)

    def test_validate_binding(self):
        schema = {
            "type": {
                reserved_key.bind: "roles"
            },
            "event": {
                reserved_key.bind: "location"
            },
            "email": {
                reserved_key.bind_regex: "[0-9]\\"
            },
            "__binder__": 4
        }

        actual_logs = self.__validate(schema, validation=ValidateSchemaBindings)
        expected_logs = [
            ml.missing_binding("roles"),
            ml.missing_binding("location"),
            ml.invalid_regex_binding(f"email.{reserved_key.bind_regex}", is_valid_regex(schema.get('email').get(reserved_key.bind_regex))),
        ]

        self.__assert_result(actual_logs, expected_logs, 3)

    def test_validate_value_type(self):
        schema = {
            "__required__": {},
            "__data_type__": 0,
            "__min_length__": True,
            "__max_length__": 0.0,
            "__min_value__": {},
            "__max_value__": [],
            "__allow_space__": 0,
            "__bind__": False,
            "__bypass__": "",
            "company": {
                "__required__": []
            }
        }

        actual_logs = self.__validate(schema, validation=ValidateValueType)
        expected_logs = [
            ml.invalid_value_type(reserved_key.required, data_type_cls.bool, data_type_cls.object),
            ml.invalid_value_type(reserved_key.data_type, data_type_cls.string, data_type_cls.integer),
            ml.invalid_value_type(reserved_key.min_length, data_type_cls.integer, data_type_cls.bool),
            ml.invalid_value_type(reserved_key.max_length, data_type_cls.integer, data_type_cls.float),
            ml.invalid_value_type(reserved_key.min_value, data_type_cls.float, data_type_cls.object),
            ml.invalid_value_type(reserved_key.max_value, data_type_cls.float, data_type_cls.array),
            ml.invalid_value_type(reserved_key.allow_space, data_type_cls.bool, data_type_cls.integer),
            ml.invalid_value_type(reserved_key.bind, data_type_cls.string, data_type_cls.bool),
            ml.invalid_value_type(reserved_key.bypass, data_type_cls.bool, data_type_cls.string),
            ml.invalid_value_type(f"company.{reserved_key.required}", data_type_cls.bool, data_type_cls.array),
        ]

        self.__assert_result(actual_logs, expected_logs, 10)

    def test_validate_keys_combinations(self):
        schema = {
            "__data_type__": "integer",
            "name": {
                "__min_value__": 4,
                "__max_value__": 10
            },
            "age": {
                "__data_type__": "integer",
                "__allow_space__": True,
                "__min_length__": 3,
                "__max_length__": 5,
                "__case__": "__lower__"
            },
            "__defaults__": {
                "__allow_space__": False,
                "__min_length__": 2,
                "__max_length__": 10,
                "__min_value__": 1,
                "__max_value__": 5,
                "__case__": "__upper__",
                "hello": {}
            }
        }

        actual_logs = self.__validate(schema, validation=ValidateKeysCombinations)
        expected_logs = [
            ml.key_support_with_object_and_array("name"),
            ml.key_support_with_object_and_array("age"),
            ml.key_support_with_number(f"name.{reserved_key.min_value}"),
            ml.key_support_with_number(f"name.{reserved_key.max_value}"),
            ml.key_support_with_string(f"age.{reserved_key.allow_space}"),
            ml.key_support_with_string(f"age.{reserved_key.min_length}"),
            ml.key_support_with_string(f"age.{reserved_key.max_length}"),
            ml.key_support_with_string(f"age.{reserved_key.case}"),
            ml.key_support_for_defaults(f"{reserved_key.defaults}.hello"),
        ]

        self.__assert_result(actual_logs, expected_logs, 9)

    def test_validate_min_max_value(self):
        schema = {
            "__min_length__": 3,
            "__max_length__": 2,
            "id": {
                "__min_length__": -1,
                "__max_length__": 10
            },
            "age": {
                "__data_type__": "integer",
                "__min_value__": 5,
                "__max_value__": 1
            },
            "__defaults__": {
                "__min_length__": -2,
                "__max_length__": -4,
                "__min_value__": 10,
                "__max_value__": 4
            }
        }

        actual_logs = self.__validate(schema, validation=ValidateMinMaxValue)
        expected_logs = [
            ml.invalid_min_max_value_or_length(reserved_key.min_length, reserved_key.max_length),
            ml.negative_min_length(f"id.{reserved_key.min_length}"),
            ml.invalid_min_max_value_or_length(f"age.{reserved_key.min_value}", f"age.{reserved_key.max_value}"),
            ml.negative_min_length(f"{reserved_key.defaults}.{reserved_key.min_length}"),
            ml.negative_or_zero_max_length(f"{reserved_key.defaults}.{reserved_key.max_length}"),
            ml.invalid_min_max_value_or_length(f"{reserved_key.defaults}.{reserved_key.min_length}", f"{reserved_key.defaults}.{reserved_key.max_length}"),
            ml.invalid_min_max_value_or_length(f"{reserved_key.defaults}.{reserved_key.min_value}", f"{reserved_key.defaults}.{reserved_key.max_value}"),
        ]

        self.__assert_result(actual_logs, expected_logs, 7)

    def test_validate_text_case(self):
        schema = {
            "first_name": {
                "__case__": "upper"
            },
            "last_name": {
                "__case__": "lower"
            },
            "full_name": {
                "__case__": "title"
            },
            "location": {
                "__data_type__": "object",
                "address": {
                    "__data_type__": "object",
                    "country": {
                        "__case__": "bad"
                    }
                }
            },
            "__defaults__": {
                "__case__": "unknown"
            }
        }

        actual_logs = self.__validate(schema, validation=ValidateTextCase)
        expected_logs = [
            ml.invalid_text_case(f"first_name.{reserved_key.case}", "upper"),
            ml.invalid_text_case(f"last_name.{reserved_key.case}", "lower"),
            ml.invalid_text_case(f"full_name.{reserved_key.case}", "title"),
            ml.invalid_text_case(f"location.address.country.{reserved_key.case}", "bad"),
            ml.invalid_text_case(f"{reserved_key.defaults}.{reserved_key.case}", "unknown"),
        ]

        self.__assert_result(actual_logs, expected_logs, 5)

    def __assert_result(self, actual_logs: list, expected_logs: list, expected_size: int = -1):

        if expected_size > -1:
            assert len(actual_logs) == expected_size

        for log in expected_logs:
            assert log in actual_logs

    def __validate(self, schema, document=None, validation: object = None):

        if document is None:
            document = {}

        output = validate(schema, document)[0].schema_result

        if validation is not None:
            output = [result for result in output if type(result.validation) is validation]

        logs = [result.message for result in output]
        return logs



from utils.util import converted_type


def test_converted_type():
    assert converted_type("") == "string"
    assert converted_type(1) == "integer"
    assert converted_type(1.0) == "float"
    assert converted_type(True) == "bool"
    assert converted_type({}) == "object"
    assert converted_type([""]) == "string_array"
    assert converted_type([1]) == "integer_array"
    assert converted_type([1.0]) == "float_array"
    assert converted_type([False]) == "bool_array"
    assert converted_type([{}]) == "object_array"
    assert converted_type([{}, 1]) == "array"
    assert converted_type(["", 1, False]) == "array"
    assert converted_type(["", 1, 0.2]) == "array"
    assert converted_type(["", 1, 0.2, {}, False]) == "array"
    assert converted_type(["", "", 0.2, {}, False]) == "array"
    assert converted_type([[]]) == "array"



import json
import os.path
import re

from jsvl.config import enable_validation_source, configs
from jsvl.models.result import Success, Info, Warn, Error
from jsvl.utils.logger import logger


class __DataType:
    def __init__(self):
        self.string = "string"
        self.integer = "integer"
        self.float = "float"
        self.bool = "bool"
        self.object = "object"
        self.object_array = "object_array"
        self.string_array = "string_array"
        self.integer_array = "integer_array"
        self.float_array = "float_array"
        self.bool_array = "bool_array"
        self.array = "array"

    def available_types(self) -> set:
        return {self.string,
                self.integer,
                self.float,
                self.bool,
                self.object,
                self.object_array,
                self.string_array,
                self.integer_array,
                self.float_array,
                self.bool_array,
                self.array}


data_type_cls = __DataType()


class ReservedKey:

    def __init__(self):
        self.required = "__required__"
        self.data_type = "__data_type__"
        self.bind = "__bind__"
        self.bind_regex = "__bind_regex__"
        self.regex_error_message = "__rem__"
        self.allow_space = "__allow_space__"
        self.min_length = "__min_length__"
        self.max_length = "__max_length__"
        self.min_value = "__min_value__"
        self.max_value = "__max_value__"
        self.case = "__case__"
        self.upper = "__upper__"
        self.lower = "__lower__"
        self.title = "__title__"
        self.bypass = "__bypass__"
        self.binder = "__binder__"
        self.defaults = "__defaults__"

    def all_keys(self) -> dict:
        return {self.required: bool,
                self.data_type: str,
                self.bind: str,
                self.bind_regex: str,
                self.regex_error_message: str,
                self.allow_space: bool,
                self.min_length: int,
                self.max_length: int,
                self.min_value: float,
                self.max_value: float,
                self.case: str,
                self.bypass: bool,
                self.binder: dict,
                self.defaults: dict}


reserved_key = ReservedKey()

regex_keys = {
    "__email__": "[^@]+@[^@]+\.[^@]+",
    "__alpha__": "^[a-zA-Z]+$",
    "__numeric__": "^([\d]+)$",
    "__alphanumeric__": "[^A-Za-z0-9]+",
    "__ipv4__": "^(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$",
    "__ipv6__": "^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))$"
}

def converted_type(field):
    field_type = field if type(field) is type else type(field)
    if field_type is str:
        return data_type_cls.string
    elif field_type is int:
        return data_type_cls.integer
    elif field_type is float:
        return data_type_cls.float
    elif field_type is bool:
        return data_type_cls.bool
    elif field_type is dict:
        return data_type_cls.object
    elif field_type is list and len(field) > 0:
        first_item_type = type(field[0])

        if first_item_type is list:
            return data_type_cls.array

        for i in field:
            if type(i) is not first_item_type:
                return data_type_cls.array

        if first_item_type is dict:
            return data_type_cls.object_array
        elif first_item_type is str:
            return data_type_cls.string_array
        elif first_item_type is int:
            return data_type_cls.integer_array
        elif first_item_type is float:
            return data_type_cls.float_array
        elif first_item_type is bool:
            return data_type_cls.bool_array
    elif field_type is list:
        return data_type_cls.array


def remove_reserved_keys(obj):
    return [k for k in obj.keys() if k not in reserved_key.all_keys().keys()]


def is_find_data_type(source: str, target: list) -> bool:
    if type(source) is str:
        for dt in source.split("|"):
            if dt in target:
                return True

    return False

def combine(path: str, key: str) -> str:
    full_path = f"{path}.{key}"
    full_path = full_path if len(path) > 0 else full_path[1:]
    return full_path[1:] if full_path.startswith(".") else full_path

def is_exact_match_data_type(source: str, target: list) -> bool:
    for dt in source.split("|"):
        if dt not in target:
            return False

    return True


def is_valid_regex(value: str):
    try:
        re.compile(value)
        return None
    except re.error as e:
        return e

def matching_data_type(key, dt, target):
    return dt == data_type_cls.array and dt in converted_type(target[key]) or {
        converted_type(target[key])}.intersection(dt.split("|"))


def is_valid_text_case(case) -> bool:
    return case in [reserved_key.upper, reserved_key.lower, reserved_key.title, None]

def read_file(file_path):
    with open(file_path, "r") as file:
        return file.read()

def is_valid_file(file_path: str) -> bool:
    return os.path.isfile(str(file_path))


def is_valid_dir(dir_path: str) -> bool:
    return os.path.isdir(str(dir_path))


def is_valid_json(source) -> bool:
    try:
        if type(source) is not dict and type(source) is not list:
            json.loads(source)
        return True
    except:
        return False


def get_as_json(source):
    if type(source) is dict or type(source) is list:
        return source
    else:
        return json.loads(source)


def dump_log(output_list: list):

    for output in output_list:
        print()

        results = []
        results.extend(output.schema_result)
        results.extend(output.document_result)

        for result in results:
            msg = result.message
            if configs.get(enable_validation_source) is True:
                msg = f"{msg} (bold)(black){type(result.validation).__name__}(end)"

            result_type = type(result)

            if result_type is Success:
                logger.success(msg)

            elif result_type is Info:
                logger.info(msg)

            elif result_type is Warn:
                logger.warn(msg)

            elif result_type is Error:
                logger.error(msg)

        print()
        print("=-" * 50)


from jsvl.config import configs, enable_output_tags, formatted_output



class AppLogger:

    def __init__(self):
        self.error_set = set()
        self.warn_set = set()
        self.success_set = set()
        self.info_set = set()
        self.all_error_list = []
        self.all_warn_list = []
        self.all_info_list = []
        self.all_success_list = []

    def warn(self, message, line_separator=False):
        output = self.get_output("(yellow)", "WARN    ::  ", message)
        formatted_message = self.apply_formatting(output)
        self.__log(formatted_message, self.warn_set, message, line_separator)

    def error(self, message, line_separator=False):
        output = self.get_output("(red)", "ERROR   ::  ", message)
        formatted_message = self.apply_formatting(output)
        self.__log(formatted_message, self.error_set, message, line_separator)

    def success(self, message, line_separator=False):
        output = self.get_output("(green)", "SUCCESS ::  ", message)
        formatted_message = self.apply_formatting(output)
        self.__log(formatted_message, self.success_set, message, line_separator)

    def info(self, message, line_separator=False):
        output = self.get_output("(blue)", "INFO    ::  ", message)
        formatted_message = self.apply_formatting(output)
        self.__log(formatted_message, self.info_set, message, line_separator)

    def get_output(self, color, tag: str, message: str) -> str:
        output = tag if configs.get(enable_output_tags) else ""
        end = "(end)" if configs.get(enable_output_tags) else ""
        return f"{color}{output}{message}{end}"

    def apply_formatting(self, output: str) -> str:
        return self.__format_output(output) if configs.get(formatted_output) else self.__remove_formatting(output)

    def line_separator(self):
        if configs.get(formatted_output):
            pass #print("" * 100)

    def has_no_error(self):
        return len(self.error_set) == 0

    def finalize(self):
        #print("*=" * 100)

        self.all_error_list.extend(self.error_set)
        self.all_warn_list.extend(self.warn_set)
        self.all_info_list.extend(self.info_set)
        self.all_success_list.extend(self.success_set)

        self.error_set.clear()
        self.warn_set.clear()
        self.info_set.clear()
        self.success_set.clear()

    def __log(self, formatted_message, log_set: set, message, line_separator=False):

        print(formatted_message)
        #prev_count = len(log_set)
        #log_set.add(message)
        #if prev_count != len(log_set):
         #   print(formatted_message)
          #  if line_separator:
           #     self.line_separator()

    def __format_output(self, text: str) -> str:
        return (text
                .replace('(bold)', "\033[1m")
                .replace("(black)", "\033[30m")
                .replace("(red)", "\033[31m")
                .replace("(green)", "\033[32m")
                .replace("(blue)", "\033[34m")
                .replace("(yellow)", "\033[33m")
                .replace("(end)", "\033[0m")
                )

    def __remove_formatting(self, text: str) -> str:
        return (text
                .replace('(bold)', "")
                .replace("(black)", "")
                .replace("(red)", "")
                .replace("(green)", "")
                .replace("(blue)", "")
                .replace("(yellow)", "")
                .replace("(end)", "")
                )


logger = AppLogger()



from abc import abstractmethod

from jsvl.utils.util import reserved_key, data_type_cls, converted_type, is_find_data_type, is_exact_match_data_type, \
    is_valid_text_case, is_valid_regex, regex_keys
import jsvl.models.schema as schema_model
from jsvl.utils.message_list import ml
from jsvl.validations.validation import Validation


class SchemaValidation(Validation):

    @abstractmethod
    def validate(self, key, schema, path):
        pass

    def is_default_object(self, key, schema, path):
        return path == reserved_key.defaults and key == reserved_key.defaults and type(schema.get(key).val) is dict

class ValidateDataType(SchemaValidation):

    def validate(self, key, schema, path):

        # hold the constraint that needs to performed
        value = schema.get(key)

        if type(value.val) is dict and key not in [reserved_key.defaults, reserved_key.binder]:
            self.__apply_validation(value.data_type, f"{path}.{reserved_key.data_type}")

        if key == reserved_key.data_type:
            self.__apply_validation(value.val, path)

    def __apply_validation(self, data_type, path):
        if type(data_type) is str:
            for dt in data_type.split("|"):
                if len(dt) == 0:
                    msg = ml.empty_value_in_data_type(path)
                    self.create_error(msg)
                elif dt not in data_type_cls.available_types():
                    msg = ml.found_invalid_data_type(path, dt)
                    self.create_error(msg)
        else:
            msg = ml.found_invalid_data_type(path)
            self.create_error(msg)


class ValidateSchemaBindings(SchemaValidation):

    def validate(self, key, schema, path):

        value = schema.get(key)
        binding = value.binding
        regex_binding = value.regex_binding

        if type(binding) is str:
            binder = schema_model.schema_doc.get("__binder__")

            if binder is None:
                self.create_error(ml.missing_binder_object())
                return

            if binder.get(binding) is None:
                self.create_error(ml.missing_binding(binding))

        if type(regex_binding) is str:
            regex_binding = self.resolve_pattern(regex_binding)
            regex_result = is_valid_regex(regex_binding)
            if regex_result is not None:
                self.create_error(ml.invalid_regex_binding(f"{path}.{reserved_key.bind_regex}", regex_result))

    def resolve_pattern(self, regex):
        return regex_keys.get(regex) if regex in regex_keys.keys() else regex


class ValidateValueType(SchemaValidation):

    def validate(self, key, schema, path):
        schema_obj = schema.get(key)

        if key not in reserved_key.all_keys().keys() and type(schema_obj.val) is not dict:
            self.create_error(f"{path} type should be {object} type")

        if type(schema_obj.val) is dict and type(schema_obj.val.get(reserved_key.data_type)) is str:
            for child_key in schema_obj.val.keys():

                if child_key in reserved_key.all_keys().keys():

                    value = schema_obj.val.get(child_key)

                    if child_key in [reserved_key.max_length, reserved_key.max_value, reserved_key.case] and value is None:
                        continue

                    if child_key in [reserved_key.min_value, reserved_key.max_value] and type(value) is int:
                        value = float(value)

                    if type(value) is not reserved_key.all_keys().get(child_key):
                        self.create_error(ml.invalid_value_type(f"{path}.{child_key}", converted_type(
                            reserved_key.all_keys().get(child_key)), converted_type(value)))

                    """
                     if value_type is not int and child_key != reserved_key.min_value and child_key != reserved_key.max_value:
                        if value_type is not reserved_key.all_keys().get(child_key):
                            self.create_error(ml.invalid_value_type(f"{path}.{child_key}", converted_type(
                                reserved_key.all_keys().get(child_key)), converted_type(schema_obj.val.get(child_key))))

                    if value_type is float and not is_find_data_type(schema_obj.data_type, [data_type_cls.float]):
                        self.create_error(ml.invalid_value_type(f"{path}.{child_key}", schema_obj.data_type,
                                                                converted_type(value_type)))

                    if value_type is int and not is_find_data_type(schema_obj.data_type, [data_type_cls.integer]):
                        self.create_error(ml.invalid_value_type(f"{path}.{child_key}", schema_obj.data_type,
                                                                converted_type(value_type)))
                    """

        if key in [reserved_key.max_length, reserved_key.max_value] and schema_obj.val is None:
            return

        if key in reserved_key.all_keys().keys():
            if type(schema_obj.val) is not reserved_key.all_keys().get(key):
                self.create_error(ml.invalid_value_type(f"{path}", converted_type(
                    reserved_key.all_keys().get(key)), converted_type(schema_obj.val)))

        if key == reserved_key.defaults and type(schema_obj.val) is dict:
            for child_key in schema_obj.val.keys():
                if child_key in [reserved_key.allow_space, reserved_key.min_length, reserved_key.max_length, reserved_key.min_value, reserved_key.max_value, reserved_key.case]:
                    value = schema_obj.val.get(child_key)

                    if child_key in [reserved_key.max_length, reserved_key.max_value] and value is None:
                        return

                    if child_key in [reserved_key.min_value, reserved_key.max_value] and type(value) is int:
                        value = float(value)

                    if type(value) is not reserved_key.all_keys().get(child_key):
                        self.create_error(ml.invalid_value_type(f"{path}.{child_key}", converted_type(
                            reserved_key.all_keys().get(child_key)), converted_type(value)))


class ValidateKeysCombinations(SchemaValidation):

    def validate(self, key, schema, path):
        schema_obj = schema.get(key)
        val = schema_obj.val

        if type(val) is dict and key != reserved_key.defaults:

            data_type = schema_obj.data_type

            # check irrelevant keys when data type is not string
            if not is_find_data_type(data_type, [data_type_cls.string, data_type_cls.array, data_type_cls.string_array,
                                                 data_type_cls.integer_array, data_type_cls.float_array,
                                                 data_type_cls.bool_array, data_type_cls.object_array]):

                string_type_support_keys = [reserved_key.min_length, reserved_key.max_length, reserved_key.allow_space,
                                            reserved_key.case]

                for string_support_key in string_type_support_keys:
                    if string_support_key in val.keys():
                        self.create_error(ml.key_support_with_string(f"{path}.{string_support_key}"))

            # check irrelevant keys when data type is not integer or float
            if not is_find_data_type(data_type, [data_type_cls.integer, data_type_cls.float]):
                number_type_support_keys = [reserved_key.min_value, reserved_key.max_value]
                for number_support_key in number_type_support_keys:
                    if number_support_key in val.keys():
                        self.create_error(ml.key_support_with_number(f"{path}.{number_support_key}"))

            # check irrelevant keys when data type is not object or object_array
            if not is_find_data_type(data_type,
                                     [data_type_cls.object, data_type_cls.object_array, data_type_cls.array]):
                for i_key in val.keys():
                    if i_key not in reserved_key.all_keys().keys():
                        self.create_error(ml.key_support_with_object_and_array(f"{path}.{i_key}"))

        # check irrelevant keys when data type is not object or object_array
        if key == reserved_key.data_type and type(val) is str:
            array_and_object_data_types = [data_type_cls.object, data_type_cls.object_array, data_type_cls.array]
            if not is_exact_match_data_type(val, array_and_object_data_types):
                for i_key in schema.keys():
                    if i_key not in reserved_key.all_keys().keys():
                        self.create_warn(ml.key_support_with_object_and_array(f"{i_key}"))

        # check if keys is default then check irrelevant keys
        if type(val) is dict and key == reserved_key.defaults:
            for child_key in val.keys():
                if child_key not in [reserved_key.allow_space, reserved_key.min_length, reserved_key.max_length, reserved_key.min_value, reserved_key.max_value, reserved_key.case]:
                    self.create_error(ml.key_support_for_defaults(f"{path}.{child_key}"))


class ValidateMinMaxValue(SchemaValidation):

    def validate(self, key, schema, path):

        schema_obj = schema.get(key)
        val = schema_obj.val

        # checking on schema root object
        # __min_length__ cannot be negative
        if key == reserved_key.min_length:
            if type(schema_obj.val) is int and schema_obj.val < 0:
                self.create_error(ml.negative_min_length(f"{reserved_key.min_length}"))

        # __max_length__ cannot be negative
        if key == reserved_key.max_length:
            if type(schema_obj.val) is int and schema_obj.val <= 0:
                self.create_error(ml.negative_or_zero_max_length(f"{reserved_key.max_length}"))

        # checking __min_length__ and __max_length__
        if reserved_key.min_length in schema.keys() and reserved_key.max_length in schema.keys():
            if type(schema_obj.val) is int and type(schema_obj.val) is int:
                if schema.get(reserved_key.min_length).val >= schema.get(reserved_key.max_length).val:
                    self.create_error(
                        ml.invalid_min_max_value_or_length(f"{reserved_key.min_length}", f"{reserved_key.max_length}"))

        # checking on schema child object and nested child object
        if type(val) is dict:

            # __min_length__ cannot be negative
            if reserved_key.min_length in val.keys():
                if type(schema_obj.min_length) is int and schema_obj.min_length < 0:
                    self.create_error(ml.negative_min_length(f"{path}.{reserved_key.min_length}"))

            # __max_length__ cannot be negative
            if reserved_key.max_length in val.keys():
                if type(schema_obj.max_length) is int and schema_obj.max_length <= 0:
                    self.create_error(ml.negative_or_zero_max_length(f"{path}.{reserved_key.max_length}"))

            # checking __min_length__ and __max_length__
            if reserved_key.min_length in val.keys() and reserved_key.max_length in val.keys():
                if type(schema_obj.min_length) is int and type(schema_obj.max_length) is int:
                    if schema_obj.min_length >= schema_obj.max_length:
                        self.create_error(ml.invalid_min_max_value_or_length(f"{path}.{reserved_key.min_length}",
                                                                             f"{path}.{reserved_key.max_length}"))

            # checking __min_value and __max_value__
            if reserved_key.min_value in val.keys() and reserved_key.max_value in val.keys():
                if type(schema_obj.min_value) is int and type(schema_obj.max_value) is int:
                    if schema_obj.min_value >= schema_obj.max_value:
                        self.create_error(ml.invalid_min_max_value_or_length(f"{path}.{reserved_key.min_value}",
                                                                             f"{path}.{reserved_key.max_value}"))


class ValidateTextCase(SchemaValidation):

    def validate(self, key, schema, path):

        obj = schema.get(key)

        if self.is_default_object(key, schema, path):
            if not is_valid_text_case(obj.case):
                self.create_error(ml.invalid_text_case(f"{path}.{reserved_key.case}", obj.case))
        elif not is_valid_text_case(obj.case):
            self.create_error(ml.invalid_text_case(f"{path}.{reserved_key.case}", obj.case))


# schema validation set
schema_validation_set = {
    ValidateDataType(),
    ValidateSchemaBindings(),
    ValidateValueType(),
    ValidateKeysCombinations(),
    ValidateMinMaxValue(),
    ValidateTextCase()
}



from abc import ABC, abstractmethod

from ordered_set import OrderedSet

class Result(ABC):

    def __init__(self, message, validation: object = None):
        self.message = message
        self.validation = validation


class Success(Result):
    pass


class Info(Result):
    pass


class Warn(Result):
    pass


class Error(Result):
    pass


class Output:

    def __init__(self, schema_result: OrderedSet, document_result: OrderedSet):
        self.schema_result = schema_result
        self.document_result = document_result



import pathlib

from setuptools import setup, find_packages

# The directory containing this file
here = pathlib.Path(__file__).parent

# The text of the README file
README = (here/"README.md").read_text()
github_url = "https://github.com/syedowaisali/json-validator-python"

setup(
    name="jsvl",
    version="1.0.7",
    url=github_url,
    project_urls={
        'Issues': f'{github_url}/issues',
        'Source Code': github_url,
    },
    author="Syed Owais Ali",
    author_email="dp.owaisali@gmail.com",
    descp="JSON Validator is a feature-rich Python library designed to elevate the validation of JSON documents by offering an extensive schema-based validation system.",
    long_description=README,
    long_description_content_type="text/markdown",
    keywords="python json schema validation",
    packages=find_packages(),
    install_requires=[
        'ordered_set',
    ],
    #py_modules=['main', 'config', 'core', 'models', 'samples', 'tests', 'utils', 'validations'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    license="GPL",
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "jsvl = jsvl.cli:run_validation",
        ],
    },
)



root_object_path = "root_object_path"
schema_file_postfix = "schema_file_postfix"
enable_output_tags = "enable_output_tags"
formatted_output = "formatted_output"
enable_output_logs = "enable_output_logs"
enable_validation_source = "enable_validation_source"
allow_space = "allow_space"
min_length = "min_length"
max_length = "max_length"
min_value = "min_value"
max_value = "max_value"
case = "case"

configs = {

    # root object path
    root_object_path: "root object",

    # schema file postfix
    schema_file_postfix: "_schema",

    # control output tags
    enable_output_tags: True,

    # control output formatting
    formatted_output: True,

    # control output logs
    enable_output_logs: True,

    # show the source filter name at end of on each log
    enable_validation_source: False,

    # set default value of allow space on any text
    allow_space: True,

    # set default minimum length
    min_length: 0,

    # set default maximum length
    max_length: None,

    # set default minimum value
    min_value: 0,

    # set default maximum value
    max_value: None,

    # set default uppercase constraint
    case: None
}



from ordered_set import OrderedSet

from jsvl.models.result import Result


class ObjectSet:

    def __init__(self):
        self.ordered_set = OrderedSet()

    def add(self, val: Result):
        messages = [result.message for result in self.ordered_set]
        if val.message not in messages:
            self.ordered_set.add(val)

    def item_set(self):
        return self.ordered_set

    def ini(self):
        self.ordered_set = OrderedSet()
